local announcement_dialog = ...

local vars = DroidButtons.vars
local ui = DroidButtons.ui
local db = DroidButtons

local bool_map  = {[true] = 'ON', [false] = 'OFF'}
local state_map = {ON = true,  OFF = false}
local val_map   = {[true] = 1, [false] = 0}

local list2 = dofile("templates.lua")

local type_radio_list = {'button', 'field'}
for i, v in ipairs(type_radio_list) do type_radio_list[v] = i end

local function px(i) return tonumber(HUDSize(i/100, 0):match("(%d+)x")) end
local function py(i) return tonumber(HUDSize(0, i/100):match("x(%d+)")) end

local function check_number(i, min, max)
	i = tonumber(i); min = tonumber(min); max = tonumber(max)
	if     (not i)           then return false 
	elseif (min and i < min) then return false
	elseif (max and i > max) then return false
	end
	return true
end

local function ispluscommand(str) 
	assert(type(str) == "string", "bad argument #1 to 'ispluscommand' (string expected, got "..type(str)..")")
	if (str:sub(1,1) == "+") then return true end
	return false
end

local function make_dirty(ctl, num)
	if (num) then 
		local result = check_number(num, 0)
		if (result) then ctl.fgcolor = "255 255 255"; vars.TabState.BadValue = false
		else             ctl.fgcolor = "255 0 0"; vars.TabState.BadValue = true
		end
	end
	ui.Tabs:MakeDirty()
end


local function create_subdlg(ctrl)
	local dlg = iup.dialog{ctrl, border="NO", menubox="NO", resize="NO", size = "FULL", bgcolor="0 0 0 0 *",}
	dlg.visible = "YES"
	return dlg
end

local function create_scrollbox(content)
	local scrollbox = iup.stationsublist{{}, control = "YES", expand = "YES",}
	iup.Append(scrollbox, create_subdlg(content))
	scrollbox[1] = 1
	return scrollbox
end






local function show_placement_interface(buttonindex, title, rtype, xpos, ypos, width, height, font)
	-- only xpos and ypos are control references, the rest should be string/number values
	local buttondef, modified
	
	local save, save_region, save_func
	local cancel, cancel_region, cancel_func
	local place_region, place_func

	-- Odd region z-order behavior allows activating multiple regions in a stack with multi-touch.
	-- Rather than destroying all regions when touch-to-place is activated, I'm just adding another region
	-- directly below the place_region to prevent the most common form of multi-touch.
	local double_touch_preventer
	
	local button_size = string.format("%dx%d", db.defaults["button_width"], db.defaults["button_height"])
	save = iup.stationbutton{title = 'Save', size = button_size, font = db.defaults["font"], cx = 8, cy = 8}
	cancel = iup.stationbutton{title = 'Cancel', size = button_size, font = db.defaults["font"], cx = tonumber(save.cx) + db.defaults["button_width"] + 8, cy = 8}

	-- Layout ---------------------------------------------------------
	iup.Append(iup.GetNextChild(db.cbox), save)
	iup.Append(iup.GetNextChild(db.cbox), cancel)
	iup.Refresh(db.cbox)
	
	-- ----------------------------------------------------------------
	
	function place_func(pW, pH)
		buttondef.x = math.floor(pW * db.xres)
		buttondef.y = math.floor(pH * db.yres)
		for axis, leg in pairs({x = 'w', y = 'h'}) do
			-- adjust axis by half the leg to center the button
			buttondef[axis] = buttondef[axis] - math.floor(buttondef[leg]/2)
			-- bring axis back within the screen if necessary
			if (buttondef[axis] < 0) then 
				buttondef[axis] = 0
			elseif (buttondef[axis] + buttondef[leg] - 1 > db[axis..'res']) then 
				buttondef[axis] = db[axis..'res'] - buttondef[leg] + 1
			end
			-- snap to grid if applicable.  I chose to always snap towards 0.
			if (db.snap_to_grid) then
				local r = buttondef[axis] % db.grid_size
				buttondef[axis] = buttondef[axis] - r
			end
		end
		modified = true
		db:UpdateButton(buttonindex, buttondef, true, false)
	end

	function cancel_func()
		HUD.help_text.title = ''

		iup.Detach(save); iup.Destroy(save); save = nil
		iup.Detach(cancel); iup.Destroy(cancel); cancel = nil

		local cleanup = Timer()
		cleanup:SetTimeout(1, function() 
			gkinterface.DestroyTouchRegion(cancel_region)
			db.regions[cancel_region] = nil
			gkinterface.DestroyTouchRegion(save_region)
			db.regions[save_region] = nil
			gkinterface.DestroyTouchRegion(place_region)
			db.regions[place_region] = nil
			gkinterface.DestroyTouchRegion(double_touch_preventer)
		end)

		db:UpdateButton(buttonindex, nil, true, false)
		ShowDialog(ui.MainDialog)
	end
	
	function save_func()
		if (modified) then 
			xpos.value = buttondef.x
			ypos.value = buttondef.y
			make_dirty()
		end
		cancel_func()
	end
	
	cancel.action = cancel_func
	save.action = save_func
	
	-- ----------------------------------------------------------------

	buttondef = {
		title = title,
		type = rtype,
		x = xpos.value,
		y = ypos.value,
		w = width,
		h = height,
		font = font,
	}
	db:UpdateButton(buttonindex, buttondef, true, false)
	cancel_region = gkinterface.CreateTouchRegion(nil, nil, nil, false, false, false, false, false, false, tonumber(cancel.cx), tonumber(cancel.cy), tonumber(cancel.cx) + db.defaults["button_width"] - 1, tonumber(cancel.cy) + db.defaults["button_height"] - 1, -100)
	save_region = gkinterface.CreateTouchRegion(nil, nil, nil, false, false, false, false, false, false, tonumber(save.cx), tonumber(save.cy), tonumber(save.cx) + db.defaults["button_width"] - 1, tonumber(save.cy) + db.defaults["button_height"] - 1, -100)
	place_region = gkinterface.CreateTouchRegion(nil, nil, nil, false, false, false, false, false, false, 0, 0, db.xres, db.yres, -99)
	double_touch_preventer = gkinterface.CreateTouchRegion(nil, nil, nil, false, false, false, false, false, false, 0, 0, db.xres, db.yres, -98)

	db.regions[cancel_region] = cancel_func
	db.regions[save_region] = save_func
	db.regions[place_region] = place_func

	HUD.help_text.title = 'Touch the screen to place the button.'
	
	HideDialog(ui.MainDialog)
end

function ui:BuildButtonEditControls()
	local buttonindex -- set by PopulateFields() below
	local newbutton -- set by PopulateFields() below
	
	local editcontrols, editcontrols_scrollbox, editcontrols_content
	local editcontrols_zbox, editcontrols_zbox_items

	-- Button region controls
	local cmd1, cmd1text, cmd1list, cmd2, cmd2text, cmd2list, cmd2c, cmd2ctext, cmd2clist
	-- Field region controls
	local xaxiscmd, yaxiscmd
	local fbc = {} -- Field Boolean Controls: invertX, invertY, mouselookX, mouselookY, relative, nonlinear
	local type_radio, type_radio_items
	-- Common region controls
	local place
	local title, xpos, ypos, width, height, font

	-- Button region controls.  They will need to be populated later.
	cmd1, cmd1text, cmd1list = list2{list = vars.commandlist, visible_items = 10, action = make_dirty}
	cmd2, cmd2text, cmd2list = list2{list = vars.commandlist, visible_items = 10}
	cmd2c, cmd2ctext, cmd2clist = list2{list = vars.commandlist, visible_items = 10, action = make_dirty, bgcolor = '55 122 150 255 *', bginactive = "128 128 128 255 *", active = "NO"}
	
	-- Field region controls
	xaxiscmd = iup.stationsublist(vars.joycmdlist)
	yaxiscmd = iup.stationsublist(vars.joycmdlist)
	fbc.invertX = iup.stationtoggle{title = 'Invert X-axis'}
	fbc.invertY = iup.stationtoggle{title = 'Invert Y-axis'}
	fbc.mouselookX = iup.stationtoggle{title = 'Mouselook X'}
	fbc.mouselookY = iup.stationtoggle{title = 'Mouselook Y'}
	fbc.relative = iup.stationtoggle{title = 'Relative'}
	fbc.nonlinear = iup.stationtoggle{title = 'Non-Linear'}

	-- Common region controls
	title = iup.text{size = cmd1text.size}
	place = iup.stationbutton{title = 'Touch-to-Place'}
	xpos = iup.text{size = px(7.5)}
	ypos = iup.text{size = px(7.5)}
	width = iup.text{size = px(7.5)}
	height = iup.text{size = px(7.5)}
	font = iup.text{size = px(7.5)}

	-- Layout ---------------------------------------------------------
	editcontrols_zbox_items = {alignment = "NW"}
	-- Button region controls
	editcontrols_zbox_items[1] = iup.vbox{
		iup.vbox{
			iup.label{title = "Button Command(s)"},
			iup.hbox{iup.label{title = 'Short-press: '}, iup.fill{}, cmd1},
		},
		iup.hbox{iup.label{title = 'Long-press: '}, iup.fill{}, cmd2},
		iup.hbox{iup.label{title = 'Long-press cancel: '}, iup.fill{}, cmd2c},
		gap = "%3",
		expand = "NO",
	}

	-- Field region controls
	editcontrols_zbox_items[2] = iup.vbox{
		iup.hbox{
			iup.vbox{
				iup.hbox{iup.label{title = "X-axis"}, iup.fill{}},
				iup.pdasubframebg{xaxiscmd},
				fbc.invertX,
				fbc.mouselookX,
				gap = "%1"
			},
			iup.vbox{
				iup.hbox{iup.label{title = "Y-axis"}, iup.fill{}},
				iup.pdasubframebg{yaxiscmd},
				fbc.invertY,
				fbc.mouselookY,
				gap = "%1"
			},
			gap = "%10",
		},
		iup.hbox{iup.fill{}, fbc.relative, fbc.nonlinear, iup.fill{}, gap = "%5" },
		gap = "%3",
		expand = "NO",
	}

	editcontrols_zbox = iup.zbox(editcontrols_zbox_items)
	

	type_radio_items = {iup.stationradio{title = 'Button', value = "ON"}, iup.stationradio{title = 'Field'}, gap = "%2"}
	type_radio = iup.radio{iup.hbox(type_radio_items)}


	editcontrols_content = iup.vbox{
		iup.pdasubframebg{iup.hbox{iup.label{title = "Region Type: "}, type_radio, iup.fill{}}},
		iup.vbox{
			iup.vbox{
				iup.hbox{iup.label{title = 'Title: '}, title},
				editcontrols_zbox,
				iup.fill{},
				gap = "%3"
			},
			iup.vbox{
				iup.hbox{iup.fill{}, place, iup.fill{}},
				iup.hbox{
					iup.vbox{
						iup.hbox{iup.label{title = 'X Position: '}, iup.fill{}, xpos,},
						iup.hbox{iup.label{title = 'Width: '}, iup.fill{}, width,},
						iup.hbox{iup.label{title = 'Font Size: '}, iup.fill{}, font,},
						gap = "%3"
					},
					iup.vbox{
						iup.hbox{iup.label{title = 'Y Position: '}, iup.fill{}, ypos},
						iup.hbox{iup.label{title = 'Height: '}, iup.fill{}, height},
						gap = "%3"
					},
					gap = "%3",
				},
				iup.fill{},
				expand = "VERTICAL",
				gap = "%3"
			},
			gap = "%2",
			margin = "%1x%1"
		},
		gap = "%1",
	}
	editcontrols_content[1].border = string.format("%d %d %d %d", px(1), py(1), px(1), py(1)) -- left, top, right, bottom
	editcontrols_content[2].border = string.format("%d %d %d %d", px(1), py(1), px(1), py(1)) -- left, top, right, bottom

	editcontrols_scrollbox = create_scrollbox(editcontrols_content)
	editcontrols_scrollbox.active = "NO"
	editcontrols_scrollbox.bgcolor = "0 0 0 0 *"

	editcontrols = iup.pdasubsubsubframe2{
		iup.pdasubsubframebg{
			editcontrols_scrollbox,
			border = string.format("%d %d %d %d", px(0.5), py(1), px(0.5), py(1)) -- left, top, right, bottom
		},
	}
	-- ----------------------------------------------------------------

	local function reset_fields()
		xpos.fgcolor = "255 255 255"
		ypos.fgcolor = "255 255 255"
		width.fgcolor = "255 255 255"
		height.fgcolor = "255 255 255"
		font.fgcolor = "255 255 255"

		vars.TabState.BadValue = false
	end
	
	function title:action(c, after) make_dirty() end

	function xaxiscmd:action(t, i, v)
		if (i == tonumber(self.cursel)) then return iup.IGNORE end
		if (v == 1) then self.cursel = i; make_dirty() end
	end
	function yaxiscmd:action(t, i, v)
		if (i == tonumber(self.cursel)) then return iup.IGNORE end
		if (v == 1) then self.cursel = i; make_dirty() end
	end
	function fbc.invertX:action(v) make_dirty() end
	function fbc.invertY:action(v) make_dirty() end
	function fbc.mouselookX:action(v) make_dirty() end
	function fbc.mouselookY:action(v) make_dirty() end
	function fbc.relative:action(v) make_dirty() end
	function fbc.nonlinear:action(v) make_dirty() end

	function place:action()
		if (vars.TabState.BadValue) then
			PopupDialog(announcement_dialog, iup.CENTER, iup.CENTER)
			return iup.IGNORE
		end
		show_placement_interface(buttonindex, title.value, type_radio.value.title:lower(), xpos, ypos, tonumber(width.value), tonumber(height.value), tonumber(font.value))
	end
		
	do	
		local function func(self, c, after)
			if (c == 27 and after == "") then return end
			make_dirty(self, after)
		end
		xpos.action = func
		ypos.action = func
		width.action = func
		height.action = func
		font.action = func
	end
		
	cmd2:SetAction(function(self, c, after) 
		local result = ispluscommand(after)
		cmd2c:Activate(result)
		if (not result) then cmd2c:Clear() end
		make_dirty()
	end)

	function type_radio:set_value(v)
		self.value = type_radio_items[v]
		type_radio_items[v]:action()
	end

	local function radio_switch(index)
		editcontrols_zbox.value = editcontrols_zbox_items[index]
		if (newbutton) then
			width.value = db.defaults[type_radio_list[index].."_width"]
			height.value = db.defaults[type_radio_list[index].."_height"]
			if (buttonindex ~= 0) then
				local buttondef = db.buttondefs[buttonindex]
				buttondef.type = type_radio_list[index]
				buttondef.w = width.value
				buttondef.h = height.value
				db:DeleteButton(buttonindex)
				db:CreateButton(buttonindex)
			end
		end
	end
	
	type_radio_items[1].action = function(self, v) 
		radio_switch(1)
	end
		
	type_radio_items[2].action = function(self, v) 
		radio_switch(2)
	end

	function editcontrols:PopulateFields(index)
		buttonindex = index
		local default = db.defaults
		local buttondef = db.buttondefs[index]

		if (buttondef) then
			newbutton = false
		else
			buttondef = {
				type = "button", 
				x = 0, 
				y = 0, 
				w = default["button_width"], 
				h = default["button_height"], 
				font = default["font"]
			}
			reset_fields()
			if (index ~= 0) then
				db.buttondefs[index] = buttondef
				db:CreateButton(index)
			end
			newbutton = true
		end

		-- Common region controls
		title.value = buttondef.title or ""
		xpos.value = buttondef.x or 0
		ypos.value = buttondef.y or 0
		width.value = buttondef.w or default[buttondef.type.."_width"]
		height.value = buttondef.h or default[buttondef.type.."_height"]
		font.value = buttondef.font or default["font"]

		-- If it's a newbutton or a reset (index == 0) then all fields should be cleared
		-- If it's an existing button, then it shouldn't matter if the other region types controls are populated
		-- Button region controls
		cmd1text.value = buttondef.command1 or ''
		cmd2text.value = buttondef.command2 or ''
		cmd2ctext.value = buttondef.command2cancel or ''
		cmd2c:Activate(buttondef.command2cancel)

		-- Field region controls
		xaxiscmd.value = vars.joycmdlist2[buttondef.command1] or vars.joycmdlist2['NONE']
		xaxiscmd.cursel = xaxiscmd.value
		yaxiscmd.value = vars.joycmdlist2[buttondef.command2] or vars.joycmdlist2['NONE']
		yaxiscmd.cursel = yaxiscmd.value
		fbc.invertX.value = bool_map[buttondef.invertX or false]
		fbc.invertY.value = bool_map[buttondef.invertY or false]
		fbc.mouselookX.value = bool_map[buttondef.mouselookX or false]
		fbc.mouselookY.value = bool_map[buttondef.mouselookY or false]
		fbc.relative.value = bool_map[buttondef.relative or false]
		fbc.nonlinear.value = bool_map[buttondef.nonlinear or false]

		type_radio:set_value(type_radio_list[buttondef.type])
		local type_idx = type_radio_list[buttondef.type]
		for i, ctl in ipairs(type_radio_items) do
			if (type_idx == i or newbutton) then ctl.active = "YES" else ctl.active = "NO" end
		end

		-- index == 0 designates a deleted record.  Disable all controls.
		editcontrols_scrollbox.active = index == 0 and "NO" or "YES"
		editcontrols_scrollbox.bgcolor = index == 0 and "0 0 0 0 *" or "255 255 255 255 *"
		editcontrols_scrollbox.scroll = "TOP"
	end
	
	function editcontrols:Save()
		if (vars.TabState.BadValue) then return false end
		local recreate_region

		local buttondef = db.buttondefs[buttonindex]
		local newtitle = buttondef.title ~= title.value

		buttondef.type = type_radio.value.title:lower()
		buttondef.title = title.value or ''
		buttondef.x = tonumber(xpos.value)
		buttondef.y = tonumber(ypos.value)
		buttondef.w = tonumber(width.value)
		buttondef.h = tonumber(height.value)
		buttondef.font = tonumber(font.value)

		if (buttondef.type == 'button') then
			local ipc = ispluscommand(buttondef.command1 or "")
			if ((ipc and cmd1text.value ~= buttondef.command1) or (not ipc and ispluscommand(cmd1text.value))) then recreate_region = true end

			buttondef.command1 = cmd1text.value
			if (buttondef.command1 == '') then buttondef.command1 = nil end
			buttondef.command2 = cmd2text.value
			if (buttondef.command2 == '') then buttondef.command2 = nil end
			buttondef.command2cancel = cmd2ctext.value
			if (buttondef.command2cancel == '') then buttondef.command2cancel = nil end

		elseif (buttondef.type == 'field') then
			local cmd1, cmd2 = vars.joycmdlist3[tonumber(xaxiscmd.value)], vars.joycmdlist3[tonumber(yaxiscmd.value)]
			if (cmd1 == "NONE") then cmd1 = nil end
			if (cmd2 == "NONE") then cmd2 = nil end
			if ((buttondef.command1 ~= cmd1) or (buttondef.command2 ~= cmd2)) then recreate_region = true end
			buttondef.command1 = cmd1
			buttondef.command2 = cmd2

			for _, p in ipairs{"invertX", "invertY", "mouselookX", "mouselookY", "relative", "nonlinear"} do
				local nv = state_map[fbc[p].value]
				if (buttondef[p] ~= nv) then recreate_region = true end
				buttondef[p] = nv
			end
		end
		
		db:save_buttondefs()
		if (recreate_region) then
			db:DeleteButton(buttonindex)
			db:CreateButton(buttonindex)
		else
			db:UpdateButton(buttonindex)
		end

		if (newbutton or newtitle) then
			return function() ui.ButtonsTab:PopulateTab(buttonindex) end
		end

		return true
	end

	function editcontrols:Undo()
		reset_fields()
		if (newbutton) then
			-- Treat undoing a newbutton as a delete
			db:DeleteButton(buttonindex)
			table.remove(db.buttondefs, buttonindex)
			return function() ui.ButtonsTab:PopulateTab(0) end
		else
			db:UpdateButton(buttonindex, nil, true, false)
			self:PopulateFields(buttonindex)
		end
	end
	-- ----------------------------------------------------------------

	return editcontrols
end
